#include "png.h"
#include "toglimg.h"

#define PNG_BYTES_TO_CHECK (8)


GLboolean toglCheckExtensionPNG(char filename[])
{
	return toglCheckExtension(filename, ".png");
}

GLboolean toglCheckImagePNG(char *file_name)
{
	FILE	*fp;
    char    sig[PNG_BYTES_TO_CHECK];

    
    if ((fp = fopen(file_name, "rb")) == NULL) {
		return GL_FALSE;
	}

	if (fread(sig, 1, PNG_BYTES_TO_CHECK, fp) != PNG_BYTES_TO_CHECK) {
		fclose(fp);
		return GL_FALSE;
    }
	
	fclose(fp);

    if (!png_check_sig(sig, PNG_BYTES_TO_CHECK)) {
		return GL_FALSE;
    }

	return GL_TRUE;
}

GLboolean toglReadImagePNG(char filename[], TOGLimage *img)
{
    png_structp     png_ptr;
    png_infop       info_ptr;
    unsigned long   width, height;
    int             bit_depth, color_type, interlace_type;
    unsigned char   **image;
	unsigned char	*p, *pbuff;
    register unsigned int	i, y;
    
	FILE			*input_file;

	png_colorp		palette;
	int				num_palette;

	GLboolean		flag_err = GL_FALSE;


	if (toglCheckImagePNG(filename) == GL_FALSE) {
		return GL_FALSE;
	}

	if ((input_file = fopen(filename, "rb")) == NULL) {
		return GL_FALSE;
	}

	if ((png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)) == NULL) {
		fclose(input_file);
		return GL_FALSE;
	}

	if ((info_ptr = png_create_info_struct(png_ptr)) == NULL) {
 		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		fclose(input_file);
		return GL_FALSE;
	}

	png_init_io(png_ptr, input_file);
	/* PNG read file header */
    png_read_info(png_ptr, info_ptr);                       
    png_get_IHDR(png_ptr, info_ptr, &width, &height,
		&bit_depth, &color_type, &interlace_type,
		NULL, NULL);

	img->m_width = width;
	img->m_height = height;

	/* Only support bit_depth 8 */
	if (bit_depth != 8) {
 		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		fclose(input_file);
		return GL_FALSE;
	}

	switch(color_type) {
		case PNG_COLOR_TYPE_GRAY:
			img->m_components = 1;
			break;
		case PNG_COLOR_TYPE_GRAY_ALPHA:
			img->m_components = 2;
			break;
		case PNG_COLOR_TYPE_RGB:
			img->m_components = 3;
			break;
		case PNG_COLOR_TYPE_RGB_ALPHA:
			img->m_components = 4;
			break;
		case PNG_COLOR_TYPE_PALETTE:
			img->m_components = 1;
			png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette);
			break;
		default:
			png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
			fclose(input_file);
			return GL_FALSE;
			break;
	}


	if (!toglAllocImage(img)) {
 		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		fclose(input_file);
		return GL_FALSE;
	}

	if ((image = (png_bytepp)malloc(height * sizeof(png_bytep))) == NULL) {
 		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		fclose(input_file);
		return GL_FALSE;
	}
    for (i = 0; i < height; i++) {
		image[i] = (png_bytep)malloc(png_get_rowbytes(png_ptr, info_ptr));
		if (image[i] == NULL) {
			flag_err = GL_TRUE;
		}
	}

	if (flag_err) {
		for (i = 0; i < height; i++) {
			if (image[i]) {
				free(image[i]);
				image[i] = NULL;
			}
		}
		if (image) {
			free(image);
			image = NULL;
		}
 		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		fclose(input_file);
		return GL_FALSE;
	}

	/* PNG read image file */
	png_read_image(png_ptr, image);


	p = img->m_buffer;
	for (y = 0; y < height; y++) {
		pbuff = image[height - y - 1];

		for(i = 0; i < width; i++) {
			switch (img->m_components) {
				case 4:
					*(p++) = *(pbuff++);
					*(p++) = *(pbuff++);
					*(p++) = *(pbuff++);
					*(p++) = *(pbuff++);
					break;
				case 3:
					*(p++) = *(pbuff++);
					*(p++) = *(pbuff++);
					*(p++) = *(pbuff++);
					*(p++) = 255;
					break;
				case 2:
					*(p++) = *(pbuff);
					*(p++) = *(pbuff);
					*(p++) = *(pbuff++);
					*(p++) = *(pbuff++);
					break;
				case 1:
				default:
					if (color_type == PNG_COLOR_TYPE_PALETTE) {
						*(p++) = palette[*(pbuff)].red;
						*(p++) = palette[*(pbuff)].green;
						*(p++) = palette[*(pbuff++)].blue;
						*(p++) = 255;
					}
					else {
						*(p++) = *(pbuff);
						*(p++) = *(pbuff);
						*(p++) = *(pbuff++);
						*(p++) = 255;
					}
					break;
			}
		}
	}

    for (i = 0; i < height; i++) {
		free(image[i]);
	}
    free(image);

    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

	fclose(input_file);

	return GL_TRUE;
}

GLboolean toglWriteImagePNG(char filename[], TOGLimage *img)
{
	FILE            *fp;
	png_structp     png_ptr;
	png_infop       info_ptr;

	time_t			gmt;
	png_time		mod_time;
	png_text		text_ptr[10];
	static char		buff[128];

	unsigned char   **image;
	int             i, j, k;

    int             color_type;

	GLboolean		flag_err = GL_FALSE;


	if ((fp = fopen(filename, "wb")) == NULL) {
		return GL_FALSE;
	}

	if ((image = (png_bytepp)calloc(img->m_height, sizeof(png_bytep))) == NULL) {
		fclose(fp);
		return GL_FALSE;
	}
	for (j = 0; j < img->m_height; j++) {
		image[j] = (png_bytep)calloc(img->m_components * img->m_width, sizeof(png_byte));
		if (image[j] == NULL) {
			flag_err = GL_TRUE;
		}
	}

	if (flag_err) {
		for (j = 0; j < img->m_height; j++) {
			if (image[j]) {
				free(image[j]);
				image[j] = NULL;
			}
		}
		if (image) {
			free(image);
			image = NULL;
		}
		fclose(fp);
		return GL_FALSE;
	}

	for (j = 0; j < img->m_height; j++) {
		k = (img->m_height - j - 1) * img->m_width * 4;
		for (i = 0; i < img->m_width; i++) {
			switch (img->m_components) {
				case 4:
					color_type = PNG_COLOR_TYPE_RGB_ALPHA;
					image[j][i * img->m_components]     = (unsigned char)img->m_buffer[k + i * 4];
					image[j][i * img->m_components + 1] = (unsigned char)img->m_buffer[k + i * 4 + 1];
					image[j][i * img->m_components + 2] = (unsigned char)img->m_buffer[k + i * 4 + 2];
					image[j][i * img->m_components + 3] = (unsigned char)img->m_buffer[k + i * 4 + 3];
					break;
				case 3:
					color_type = PNG_COLOR_TYPE_RGB;
					image[j][i * img->m_components]     = (unsigned char)img->m_buffer[k + i * 4];
					image[j][i * img->m_components + 1] = (unsigned char)img->m_buffer[k + i * 4 + 1];
					image[j][i * img->m_components + 2] = (unsigned char)img->m_buffer[k + i * 4 + 2];
					break;
				default:
					break;
			}
		}
	}

	if ((png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)) == NULL) {
		for (j = 0; j < img->m_height; j++) {
			if (image[j]) {
				free(image[j]);
				image[j] = NULL;
			}
		}
		if (image) {
			free(image);
			image = NULL;
		}
		
		fclose(fp);
		return GL_FALSE;
	}

	if ((info_ptr = png_create_info_struct(png_ptr)) == NULL) {
 		png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);

		for (j = 0; j < img->m_height; j++) {
			if (image[j]) {
				free(image[j]);
				image[j] = NULL;
			}
		}
		if (image) {
			free(image);
			image = NULL;
		}

		fclose(fp);
		return GL_FALSE;
	}


#if 0
	if (setjmp(png_ptr->jmpbuf)) {
		png_destroy_write_struct(&png_ptr,  &info_ptr);

		for (j = 0; j < img->m_height; j++) {
			if (image[j]) {
				free(image[j]);
				image[j] = NULL;
			}
		}
		if (image) {
			free(image);
			image = NULL;
		}
		
		fclose(input_file);
		return GL_FALSE;
	}
#endif
	
	png_init_io(png_ptr, fp);
	png_set_filter(png_ptr, 0, PNG_ALL_FILTERS);
	png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
	/* 8bit, interlace none */
	png_set_IHDR(png_ptr, info_ptr, img->m_width, img->m_height,
		8, color_type, PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
	/* gamma 1.0 */
    png_set_gAMA(png_ptr, info_ptr, 1.0);


	/* Textual data */

	time(&gmt);
	png_convert_from_time_t(&mod_time, gmt);
	png_set_tIME(png_ptr, info_ptr, &mod_time);

	text_ptr[0].key = "Title";
	text_ptr[0].text = filename;
	text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[1].key = "Author";
	text_ptr[1].text = getenv("USER");
	text_ptr[1].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[2].key = "Description";
	text_ptr[2].text = "";
	text_ptr[2].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[3].key = "Copyright";
	text_ptr[3].text = "";
	text_ptr[3].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[4].key = "Creation Time";
	text_ptr[4].text = png_convert_to_rfc1123(png_ptr, &mod_time);
	text_ptr[4].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[5].key = "Software";
	sprintf(buff, "libtoglimg %s", TOGL_VERSION_IMAGE);
	text_ptr[5].text = buff;
	text_ptr[5].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[6].key = "Disclaimer";
	text_ptr[6].text = "";
	text_ptr[6].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[7].key = "Warning";
	text_ptr[7].text = "";
	text_ptr[7].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[8].key = "Source";
	text_ptr[8].text = "";
	text_ptr[8].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[9].key = "Comment";
	text_ptr[9].text = "";
	text_ptr[9].compression = PNG_TEXT_COMPRESSION_NONE;

	png_set_text(png_ptr, info_ptr, text_ptr, 10);

	/* PNG write header */
	png_write_info(png_ptr, info_ptr);
	/* PNG write image */
	png_write_image(png_ptr, image);

	png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr, &info_ptr);

	for (j = 0; j < img->m_height; j++) {
		if (image[j]) {
			free(image[j]);
			image[j] = NULL;
		}
	}
	if (image) {
		free(image);
		image = NULL;
	}

	fclose(fp);

	return GL_TRUE;
}
